import hashlib
import json
import logging
import os
import pickle
import sys
import twint

from copy import deepcopy
from datetime import datetime, timedelta, timezone
from dateutil.parser import parse as parsedate
from feedgen.feed import FeedGenerator
from time import sleep

import googleapiclient.discovery
import googleapiclient.errors
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from googleapiclient.errors import HttpError

RECHECK_INTERVAL = 20
START_DELAY = 15
RESTART_MINUTES = 60

YT_API_KEY = "apikey.json"
YT_TOKEN_FILE = "/disk1/kiki_rss/yt_token.pickle"
YT_PLAYLIST_ID = "UUqF1M9cLvoB-zJsLQgIxx4Q"
YT_SCOPES = ["https://www.googleapis.com/auth/youtube.readonly"]
YT_ATTEMPT_REAUTH = False
YT_GET_VIDEOS = 20

TW_USERNAME = "bunnyonvt"
TW_LIMIT = 20

LOG_FILE = "/disk1/kiki_rss/notifier.log"
DATA_FILE = "/disk1/kiki_rss/data.pickle"
FEED_OUTPUT = "/disk1/webserver/rabbalabs/rss/feed.xml"

FEED_PARAMS = {
    "title": "Kiki Notifications",
    "link": "https://rabbalabs.com/",
    "description": "Notifications for activity on various platforms by our favorite rabba Kiki.",
}
MAX_ENTRIES = 60
TITLE_LIMIT = 40
ENTRY_CACHE_SIZE = 200

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
formatter = logging.Formatter(
    fmt="%(asctime)s %(name)s.%(levelname)s: %(message)s", datefmt="%Y.%m.%d %H:%M:%S"
)
handler = logging.FileHandler(LOG_FILE, encoding="utf-8")
handler.setFormatter(formatter)
logger.addHandler(handler)


def get_google_auth():
    os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"
    api_service_name = "youtube"
    api_version = "v3"
    client_secrets_file = YT_API_KEY

    if os.path.exists(YT_TOKEN_FILE):
        with open(YT_TOKEN_FILE, "rb") as token:
            credentials = pickle.load(token)
    else:
        credentials = None
    if not credentials or not credentials.valid:
        try:
            credentials.refresh(Request())
        except:
            if YT_ATTEMPT_REAUTH:
                flow = InstalledAppFlow.from_client_secrets_file(YT_API_KEY, YT_SCOPES)
                credentials = flow.run_local_server(port=32104)
            else:
                raise
        with open(YT_TOKEN_FILE, "wb") as token:
            pickle.dump(credentials, token)
    return googleapiclient.discovery.build(
        api_service_name, api_version, credentials=credentials
    )


def check_youtube_videos(api):
    request = api.playlistItems().list(
        part="contentDetails",
        playlistId=YT_PLAYLIST_ID,
        maxResults=YT_GET_VIDEOS,
    )
    response = request.execute()
    video_ids = ",".join(
        [item["contentDetails"]["videoId"] for item in response["items"]]
    )
    request = api.videos().list(
        part="contentDetails, id, liveStreamingDetails, snippet",
        id=video_ids,
    )
    response = request.execute()
    video_notifications = {}
    for video in response.get("items", []):
        video_status = video.get("snippet", {}).get("liveBroadcastContent")
        if video_status == "upcoming" and video["liveStreamingDetails"].get(
            "actualStartTime"
        ):
            video_status = "live"
        if video_status and video_status != "none":
            video_type = (
                "premiere"
                if video.get("contentDetails", {}).get("duration")
                and video["contentDetails"]["duration"] != "P0D"
                else "stream"
            )
            if video_status == "live":
                start_time = f'<b>Start Time:</b> {video["liveStreamingDetails"]["actualStartTime"]}'
                publish_time = parsedate(
                    video["liveStreamingDetails"]["actualStartTime"]
                )
                hash_time = video["snippet"]["publishedAt"]
            else:
                start_time = f'<b>Scheduled Time:</b> {video["liveStreamingDetails"]["scheduledStartTime"]}'
                publish_time = parsedate(video["snippet"]["publishedAt"])
                hash_time = video["liveStreamingDetails"]["scheduledStartTime"]
            description = video["snippet"]["description"].replace("\n", "<br>")
            entry_body = f"""\
<p>{start_time}</p>
<p><b>Description:</b> {description}</p>\
"""
            video_notifications[video["id"]] = {
                "title": f'[{video_type.capitalize()} {video_status.capitalize()}]: {video["snippet"]["title"][:TITLE_LIMIT]}{"..." if video["snippet"]["title"][TITLE_LIMIT:] else ""}',
                "link": f'https://www.youtube.com/watch?v={video["id"]}',
                "category": {
                    "label": f"Youtube: {video_status.capitalize()}",
                    "term": f"yt_{video_status}",
                },
                "body": entry_body,
                "published": publish_time,
                "guid": hashlib.md5(
                    f'{video["id"]}{video_status}{hash_time}'.encode()
                ).hexdigest(),
            }
    return video_notifications


def check_twitter_posts(config):
    twitter_notifications = {}
    twint.run.Search(config)
    for tweet in twint.output.tweets_list:
        tweet_time = parsedate(f"{tweet.datestamp} {tweet.timestamp}{tweet.timezone}")
        entry_body = f"""\
<p><b>Posted:</b> {tweet_time.astimezone(timezone.utc)}</p>
<p><b>Tweet:</b> {tweet.tweet}</p>\
"""
        twitter_notifications[tweet.id] = {
            "title": f"[Twitter]: {tweet.tweet[:TITLE_LIMIT]}{'...' if tweet.tweet[TITLE_LIMIT:] else ''}",
            "link": tweet.link,
            "category": {
                "label": "Twitter",
                "term": "twitter",
            },
            "body": entry_body,
            "published": tweet_time,
            "guid": hashlib.md5(f"{tweet.id}".encode()).hexdigest(),
        }
    return twitter_notifications


def generate_feed(feed_data):
    feed = FeedGenerator()
    feed.title(FEED_PARAMS["title"])
    feed.link(href=FEED_PARAMS["link"])
    feed.description(FEED_PARAMS["description"])
    for item in sorted(feed_data.values(), key=lambda i: i["published"], reverse=True)[
        :MAX_ENTRIES
    ]:
        entry = feed.add_entry(order="append")
        entry.title(item["title"])
        entry.link(href=item["link"])
        entry.category(item["category"])
        entry.description(item["body"])
        entry.published(item["published"])
        entry.guid(item["guid"])
    feed.rss_file(FEED_OUTPUT)


def trim_feed_cache(feed_data):
    trimmed_entries = sorted(
        list(feed_data.items()), key=lambda i: i[1]["published"], reverse=True
    )[:ENTRY_CACHE_SIZE]
    trimmed_data = dict(trimmed_entries)
    return trimmed_data


def main():
    logger.info(f"Waiting {START_DELAY} seconds before starting.")
    sleep(START_DELAY)
    logger.info("Starting monitor service.")
    youtube_api = get_google_auth()
    twint_config = twint.Config()
    twint_config.Username = TW_USERNAME
    twint_config.Limit = TW_LIMIT
    twint_config.Store_object = True
    twint_config.Hide_output = True
    feed_data = {}
    logged_time = datetime.now()
    time_yt = timedelta()
    time_tw = timedelta()
    time_feed = timedelta()
    total_syncs = 0
    if os.path.exists(DATA_FILE):
        with open(DATA_FILE, "rb") as loadfile:
            feed_data = pickle.load(loadfile)
    if len(feed_data) > ENTRY_CACHE_SIZE:
        feed_data = trim_feed_cache(feed_data)
        with open(DATA_FILE, "wb") as savefile:
            pickle.dump(feed_data, savefile)
    feed_data_prev = deepcopy(feed_data)
    while True:
        time_start = datetime.now()
        try:
            yt_items = check_youtube_videos(youtube_api)
            feed_data.update(yt_items)
        except Exception as e:
            logger.error("Error syncing YouTube videos:")
            logger.exception(e)
        time_yt += datetime.now() - time_start
        time_start = datetime.now()
        try:
            tw_items = check_twitter_posts(twint_config)
            feed_data.update(tw_items)
        except Exception as e:
            logger.error("Error syncing Twitter posts:")
            logger.exception(e)
        time_tw += datetime.now() - time_start
        if feed_data != feed_data_prev:
            time_start = datetime.now()
            logger.info("Updating feed.")
            generate_feed(feed_data)
            with open(DATA_FILE, "wb") as savefile:
                pickle.dump(feed_data, savefile)
            feed_data_prev = deepcopy(feed_data)
            time_feed += datetime.now() - time_start
        total_syncs += 1
        if datetime.now() - logged_time > timedelta(minutes=RESTART_MINUTES):
            logger.info(
                f"Restarting service. Synced {total_syncs} times in {RESTART_MINUTES} minutes."
            )
            logger.info(
                f"Spent {time_yt} checking YouTube, {time_tw} checking Twitter."
            )
            if time_feed:
                logger.info(f"Spent {time_feed} updating the feed.")
            return
        sleep(RECHECK_INTERVAL)


if __name__ == "__main__":
    main()
