import json
import os
import pickle
import re
import requests
import googleapiclient.discovery
import googleapiclient.errors

from dateutil.parser import parse as parsedate
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from googleapiclient.errors import HttpError
from itertools import chain
from time import sleep

CHANNEL_ID = "UCqF1M9cLvoB-zJsLQgIxx4Q"
VIDEO_IDS_FILE = "/disk1/chat_downloader/videos.json"
GAME_PLAYLIST_FILE = "/disk1/chat_downloader/game_playlists.json"
UNTAGGED_VODS_FILE = "/disk1/chat_downloader/untagged_vods.json"
MANUAL_OVERRIDE_FILE = "playlist_maker/manual_override.json"

API_KEY = "apikey.json"
TOKEN_FILE = "/disk1/chat_downloader/get_videos_token.pickle"
SCOPES = ["https://www.googleapis.com/auth/youtube.readonly"]

YT_BATCH_SIZE = 50

verbose_output = False
save_untagged = False
attempt_reauth = False
game_info_pattern = re.compile(
    r'{"style":"RICH_METADATA_RENDERER_STYLE_BOX_ART".*?"url":"([^"]+)","width":\d+,"height":\d+}]},"title":{"simpleText":"([^"]+)"},(?:"subtitle":{"simpleText":"(\d+)")?.*?{"browseId":"([^"]+)"'
)
alt_info_pattern = re.compile(
    r'{"type":"INTERACTIVE_TABBED_HEADER_RENDERER_TYPE_GAMING","title":{"simpleText":"([^"]+)"}.*?(?:"metadata":{"runs":\[{"text":"(\d+)".*?)?"boxArt":{"thumbnails":.*?\/([^=\/]+)='
)
manual_override_substitutions = {}
game_info_cache = {}


def get_google_auth():
    os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "1"
    api_service_name = "youtube"
    api_version = "v3"
    client_secrets_file = API_KEY

    if os.path.exists(TOKEN_FILE):
        with open(TOKEN_FILE, "rb") as token:
            credentials = pickle.load(token)
    else:
        credentials = None
    if not credentials or not credentials.valid:
        try:
            credentials.refresh(Request())
        except:
            if attempt_reauth:
                flow = InstalledAppFlow.from_client_secrets_file(API_KEY, SCOPES)
                credentials = flow.run_local_server(port=32104)
            else:
                raise
        with open(TOKEN_FILE, "wb") as token:
            pickle.dump(credentials, token)
    return googleapiclient.discovery.build(
        api_service_name, api_version, credentials=credentials
    )


def get_game_data(video_id):
    if video_id in manual_override_substitutions:
        game_id = manual_override_substitutions[video_id]
        if game_id in game_info_cache:
            if verbose_output:
                print(f"Using cached game info: {game_info_cache[game_id]}")
            return game_info_cache[game_id]
        if game_id == "IGNORE":
            if verbose_output:
                print("Skipping because video ID set to ignore.")
            return None
        game_info_url = f"https://www.youtube.com/channel/{game_id}"
        if verbose_output:
            print(f"Getting {game_info_url}")
        response = requests.get(game_info_url)
        if match := alt_info_pattern.search(response.text):
            game_info = {
                "id": game_id,
                "title": match.group(1),
                "year": match.group(2) or "",
                "image": f"https://yt3.ggpht.com/{match.group(3)}=s200-w144-h200-c-k-c0x00ffffff-no-nd-rj",
            }
            game_info_cache[game_info["id"]] = game_info
            if verbose_output:
                print(f"Found {game_info}")
            return game_info
    vod_url = f"https://www.youtube.com/watch?v={video_id}"
    if verbose_output:
        print(f"Getting {vod_url}")
    response = requests.get(vod_url)
    game_info = None
    if match := game_info_pattern.search(response.text):
        game_info = {
            "id": match.group(4),
            "title": match.group(2),
            "year": match.group(3) or "",
            "image": match.group(1),
        }
        game_info_cache[game_info["id"]] = game_info
    if verbose_output:
        print(f"Found {game_info}")
    return game_info


def update_timestamps(playlists):
    item_list = list(chain(*[item["videos"] for item in playlists.values()]))
    item_dict = {item["id"]: item for item in item_list}
    youtube = get_google_auth()
    page = ""
    response_items = []
    retries = 0
    curr_index = 0
    print("Getting stream start times for videos.")
    while curr_batch := item_list[curr_index : curr_index + YT_BATCH_SIZE]:
        if verbose_output:
            print(f"Checking items {curr_index + 1} - {curr_index + YT_BATCH_SIZE}")
        try:
            request = youtube.videos().list(
                part="id, liveStreamingDetails",
                id=",".join([item["id"] for item in curr_batch]),
                maxResults=YT_BATCH_SIZE,
            )
            response = request.execute()
            curr_index += YT_BATCH_SIZE
            response_items.extend(response["items"])
        except HttpError as err:
            if retries >= MAX_RETRIES:
                raise err
            retries += 1
            print(
                f"Waiting {RETRY_DELAY} seconds to retry due to error. ({retries} of {MAX_RETRIES})"
            )
            sleep(RETRY_DELAY)
    for item in response_items:
        if actual_start_time := item.get("liveStreamingDetails", {}).get(
            "actualStartTime"
        ):
            item_dict[item["id"]]["timestamp"] = actual_start_time


def main():
    global manual_override_substitutions
    if os.path.isfile(MANUAL_OVERRIDE_FILE):
        with open(MANUAL_OVERRIDE_FILE, encoding="utf-8") as overrides_file:
            manual_override_substitutions = {
                stream_id: game_id
                for game_id, (game_title, streams) in json.load(overrides_file).items()
                for stream_id in streams
            }
    print("Getting playlist info from VODs in selected channel.")
    with open(VIDEO_IDS_FILE, encoding="utf-8") as videos_json:
        video_ids = json.load(videos_json)
    to_process = [
        video for video in video_ids if video["snippet"]["channelId"] == CHANNEL_ID
    ]
    playlists = {}
    untagged = []
    for video in to_process:
        game_info = get_game_data(video["contentDetails"]["videoId"])
        if game_info:
            if game_info["id"] not in playlists:
                playlists[game_info["id"]] = {
                    "title": game_info["title"],
                    "year": game_info["year"],
                    "image": game_info["image"],
                    "videos": [],
                }
            playlists[game_info["id"]]["videos"].append(
                {
                    "id": video["contentDetails"]["videoId"],
                    "title": video["snippet"]["title"],
                    "timestamp": video["contentDetails"]["videoPublishedAt"],
                    "thumbnail": video["snippet"]["thumbnails"]["high"]["url"],
                }
            )
        elif save_untagged:
            untagged.append(
                {
                    "id": video["contentDetails"]["videoId"],
                    "title": video["snippet"]["title"],
                    "link": f'https://www.youtube.com/watch?v={video["contentDetails"]["videoId"]}',
                }
            )
    if playlists:
        update_timestamps(playlists)
        with open(GAME_PLAYLIST_FILE, "w", encoding="utf-8") as playlists_json:
            json.dump(playlists, playlists_json, indent=4, ensure_ascii=False)
    if untagged:
        with open(UNTAGGED_VODS_FILE, "w", encoding="utf-8") as untagged_json:
            json.dump(untagged, untagged_json, indent=4, ensure_ascii=False)


if __name__ == "__main__":
    verbose_output = True
    save_untagged = True
    attempt_reauth = True
    main()
