import json
import os
import re
import requests
import sqlite3
from bs4 import BeautifulSoup
from time import sleep

DATABASE_FILE = "/disk1/pyondex/database.db"
THREAD_LIST = "/disk1/pyondex/new_threads.json"
SAVE_INTERVAL = 20
MAX_RETRIES = 5
RETRY_DELAY = 10
BASE_ARCHIVE_URL = "https://warosu.org"
# SQLite options
DB_CACHE_SIZE = 10000
DB_MMAP_SIZE = 2147483648
DB_TEMP_STORE = "MEMORY"
DB_SYNCHRONOUS = "NORMAL"

verbose_output = False

quotes_regex = re.compile(r"^(?:(?:/vt/post/S?)|(?:#p))(\d+)$")
image_regex = re.compile(r"^//i\.warosu\.org/data/vt/img(?:/\d+)+\.[a-z]+$")
hash_regex = re.compile(r"^/vt/image/([A-Za-z0-9\-_]+)$")
backlink_regex = re.compile(r'<a class="backlink" href="#p(\d+)".*?</a>')
file_info_regex = re.compile(r"^File: [\w\. ]+, \w+, .*?\.[a-zA-Z]+$")


def create_tables(db_cur):
    tables = [
        """\
CREATE TABLE IF NOT EXISTS posts (
    id                INTEGER NOT NULL PRIMARY KEY,
    timestamp         INTEGER NOT NULL,
    thread            INTEGER NOT NULL,
    deleted           INTEGER NOT NULL,
    plain_text        TEXT NOT NULL,
    html_text         TEXT NOT NULL,
    name              TEXT NOT NULL,
    tripcode          TEXT,
    subject           TEXT,
    quoting           TEXT,
    image_url         TEXT,
    image_thumbnail   TEXT,
    image_hash        TEXT,
    image_filesize    TEXT,
    image_dimensions  TEXT,
    image_filename    TEXT
)\
""",
    ]
    indexes = [
        """\
CREATE INDEX IF NOT EXISTS post_time_index
    ON posts (timestamp)\
""",
        """\
CREATE INDEX IF NOT EXISTS post_thread_index
    ON posts (thread)\
""",
        """\
CREATE INDEX IF NOT EXISTS deleted_post_index
    ON posts (deleted)\
""",
        """\
CREATE INDEX IF NOT EXISTS post_subject_index
    ON posts (subject)\
""",
        """\
CREATE INDEX IF NOT EXISTS post_name_index
    ON posts (name)\
""",
        """\
CREATE INDEX IF NOT EXISTS post_tripcode_index
    ON posts (tripcode)\
""",
        """\
CREATE INDEX IF NOT EXISTS post_image_hash_index
    ON posts (image_hash)\
""",
        """\
CREATE INDEX IF NOT EXISTS post_image_filename_index
    ON posts (image_filename)\
""",
    ]
    for create_item in [*tables, *indexes]:
        db_cur.execute(create_item)


def get_db():
    db = sqlite3.connect(DATABASE_FILE)
    db.execute(f"PRAGMA cache_size = {DB_CACHE_SIZE}")
    db.execute(f"PRAGMA mmap_size = {DB_MMAP_SIZE}")
    db.execute(f"PRAGMA temp_store = {DB_TEMP_STORE}")
    db.execute(f"PRAGMA synchronous = {DB_SYNCHRONOUS}")
    db_cur = db.cursor()
    create_tables(db_cur)
    return db, db_cur


def process_post(post, thread_id=None, prev_timestamp=None):
    post_id = int(post.attrs["id"].rpartition("p")[2])
    post_timestamp = prev_timestamp
    if (
        find_timestamp := post.find("span", {"class", "posttime"})
    ) and find_timestamp.attrs["title"]:
        post_timestamp = int(find_timestamp.attrs["title"])
    else:
        print(
            f"WARNING: Timestamp is missing from >>{post_id}, using previous timestamp of {prev_timestamp} instead."
        )
    post_deleted = int(bool(post.find("img", {"alt": "[DELETED]"})))
    post_html = post.find("p", {"itemprop": "text"})
    post_plain_text = post_html.get_text(separator=" ").strip()
    post_html_text = backlink_regex.sub(
        lambda i: f'<a href="/vt/post/{i.group(1)}">&gt;&gt;{i.group(1)}</a>',
        post_html.decode_contents().strip(),
    )
    post_user_name = post.find("span", {"itemprop": "name"}).text
    post_tripcode = None
    if find_tripcode := post.find("span", {"class": "postertrip"}):
        post_tripcode = find_tripcode.text.strip()
    post_subject = None
    if find_subject := post.find("span", {"class": "filetitle"}):
        post_subject = find_subject.text
    post_quoting = None
    if find_quotes := post_html.find_all("a", {"href": quotes_regex}):
        post_quoting = json.dumps(
            [
                quotes_regex.fullmatch(match.attrs["href"]).group(1)
                for match in find_quotes
            ]
        )
    image_url = None
    image_thumbnail = None
    image_hash = None
    image_filesize = None
    image_dimensions = None
    image_filename = None
    if find_image := post.find("a", {"href": image_regex}):
        image_url = find_image.attrs["href"]
        image_thumbnail = find_image.find("img", {"class": "thumb"}).attrs["src"]
        image_hash = hash_regex.fullmatch(
            post.find("a", {"href": hash_regex}).attrs["href"]
        ).group(1)
        if image_metadata := post.find("span", string=file_info_regex):
            (
                image_filesize,
                image_dimensions,
                image_filename,
            ) = image_metadata.text.rpartition("File: ")[2].split(", ", 2)
        else:
            print(f"WARNING: Could not extract image metadata for >>{post_id}.")
    return {
        "id": post_id,
        "timestamp": post_timestamp,
        "thread": thread_id or post_id,
        "deleted": post_deleted,
        "plain_text": post_plain_text,
        "html_text": post_html_text,
        "name": post_user_name,
        "tripcode": post_tripcode,
        "subject": post_subject,
        "quoting": post_quoting,
        "image_url": image_url,
        "image_thumbnail": image_thumbnail,
        "image_hash": image_hash,
        "image_filesize": image_filesize,
        "image_dimensions": image_dimensions,
        "image_filename": image_filename,
    }


def process_thread(thread):
    thread_url = f"{BASE_ARCHIVE_URL}{thread}"
    if verbose_output:
        print(f"Checking {thread_url}")
    retries = 0
    while True:
        try:
            thread_html = requests.get(thread_url)
            soup = BeautifulSoup(thread_html.content, "html.parser")
            thread_data = []
            op_post = soup.find(
                "div", {"itemtype": "http://schema.org/DiscussionForumPosting"}
            )
            op_data = process_post(op_post)
            prev_timestamp = op_data["timestamp"]
            thread_data.append(op_data)
            for post in soup.find_all("td", {"class": "reply"}):
                post_data = process_post(post, op_data["id"], prev_timestamp)
                prev_timestamp = post_data["timestamp"]
                thread_data.append(post_data)
            if verbose_output:
                print(f"Found {len(thread_data)} posts.")
            break
        except Exception as e:
            retries += 1
            if retries > MAX_RETRIES:
                raise e
            print(f"Error in {thread_url}, retrying in {RETRY_DELAY} seconds.")
            sleep(RETRY_DELAY)
    return thread_data


class DBUpdateQueue:
    def __init__(self, db, db_cur, db_update_interval):
        self.db = db
        self.db_cur = db_cur
        self.update_queue = []
        self.queued_posts = 0
        self.total_saved_threads = 0
        self.total_saved_posts = 0
        self.db_update_interval = db_update_interval

    def add_record(self, thread_data):
        self.queued_posts += len(thread_data)
        self.update_queue.append(thread_data)
        if len(self.update_queue) >= self.db_update_interval:
            self.save_queue()

    def save_queue(self):
        if not len(self.update_queue):
            return
        if verbose_output:
            print(
                "Saving {} post{} from {} thread{}...".format(
                    self.queued_posts,
                    "s"[: self.queued_posts ^ 1],
                    len(self.update_queue),
                    "s"[: len(self.update_queue) ^ 1],
                )
            )
        self.write_to_db()
        self.db.commit()
        self.total_saved_posts += self.queued_posts
        self.queued_posts = 0
        self.total_saved_threads += len(self.update_queue)
        self.update_queue = []

    def write_to_db(self):
        to_update = []
        for thread in self.update_queue:
            for post in thread:
                to_update.append(
                    [
                        post["id"],
                        post["timestamp"],
                        post["thread"],
                        post["deleted"],
                        post["plain_text"],
                        post["html_text"],
                        post["name"],
                        post["tripcode"],
                        post["subject"],
                        post["quoting"],
                        post["image_url"],
                        post["image_thumbnail"],
                        post["image_hash"],
                        post["image_filesize"],
                        post["image_dimensions"],
                        post["image_filename"],
                    ]
                )
        self.db_cur.executemany(
            "INSERT OR REPLACE INTO posts VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
            to_update,
        )


def main():
    print("Updating DB for stored threads.")
    db, db_cur = get_db()
    to_process = []
    if os.path.isfile(THREAD_LIST):
        with open(THREAD_LIST) as threads_json:
            to_process = json.load(threads_json)
    update_queue = DBUpdateQueue(db, db_cur, SAVE_INTERVAL)
    total_threads = len(to_process)
    for thread_num, thread in enumerate(to_process):
        thread_data = process_thread(thread)
        if thread_data:
            update_queue.add_record(thread_data)
    update_queue.save_queue()
    db.close()
    print("Done.")
    print(
        "Saved {} post{} from {} thread{}.".format(
            update_queue.total_saved_posts,
            "s"[: update_queue.total_saved_posts ^ 1],
            update_queue.total_saved_threads,
            "s"[: update_queue.total_saved_threads ^ 1],
        )
    )


if __name__ == "__main__":
    verbose_output = True
    main()
