import logging
import re
import requests
import sys
from bs4 import BeautifulSoup
from datetime import datetime, timezone
from dateutil.parser import parse as parsedate
from flask import Flask, render_template, request
from flask_bootstrap import Bootstrap4
from flask_wtf import FlaskForm
from wtforms import IntegerField, StringField, SubmitField, TextAreaField
from wtforms.validators import InputRequired

BASE_THREAD_URL = "https://boards.4channel.org/vt/thread/"
BASE_VIDEO_URL = "https://www.youtube.com/watch?v="
VALID_PREFIXES = ("`", "kt!")
MAX_4CHAN_POST_LINES = 100

URL_PREFIX = "/tagger"
LOG_FILE = "/disk1/kiki_tagger/server.log"


class SourceFilter(logging.Filter):
    def filter(self, record):
        record.request_source = (
            request.headers["Cf-Connecting-Ip"]
            if "Cf-Connecting-Ip" in request.headers
            else "Local"
        )
        return True


class RequestForm(FlaskForm):
    video_entry = StringField("Stream ID/URL", validators=[InputRequired()])
    delay_entry = IntegerField(
        "Default Offset", validators=[InputRequired()], default=15
    )
    threads_entry = TextAreaField(
        "Thread ID/URLs (one per line)", validators=[InputRequired()]
    )
    submit_request = SubmitField("Get Tags")

    class Meta:
        csrf = False


logger = logging.getLogger(__name__)
log_handler = logging.FileHandler(LOG_FILE, encoding="utf-8")
log_handler.setFormatter(
    logging.Formatter(
        fmt="%(asctime)s [%(levelname)s] (%(request_source)s) %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
)
log_handler.addFilter(SourceFilter())
logging.basicConfig(handlers=[log_handler], level=logging.INFO)

title_regex = re.compile(r'"title":{"simpleText":"(.*?)"}')
timestamp_regex = re.compile(r'"startTimestamp":"(.*?)"')
quotes_regex = re.compile(r">>\d+\s*")
offset_regex = re.compile(r"\(-(\d+)\)$")
add_delay_regex = re.compile(r"offset (\d+)")

app = Flask(__name__)
app.config["SECRET_KEY"] = ""
bootstrap = Bootstrap4(app)


def get_video_attributes(video_id):
    video_html = requests.get(f"{BASE_VIDEO_URL}{video_id}")
    title_search = title_regex.search(video_html.text)
    video_title = title_search.group(1) if title_search else None
    timestamp_search = timestamp_regex.search(video_html.text)
    video_timestamp = timestamp_search.group(1) if timestamp_search else None
    return {
        "id": video_id,
        "title": video_title,
        "timestamp": parsedate(video_timestamp),
    }


def parse_thread(thread_id):
    thread_html = requests.get(f"{BASE_THREAD_URL}{thread_id}")
    soup = BeautifulSoup(thread_html.content, "html.parser")
    tags = []
    for message in soup.find_all(class_="postContainer"):
        message_id = message["id"].lstrip("pc")
        message_time = message.find(class_="dateTime").get("data-utc")
        message_text = message.find(class_="postMessage").get_text(separator=" ")
        quote_link = message.find(class_="quotelink")
        message_quoting = quote_link.get_text().lstrip(">") if quote_link else None
        if message_quoting:
            while quote_link := quotes_regex.match(message_text):
                message_text = message_text.replace(quote_link.group(), "", 1)
        message_offset = 0
        if time_offset := offset_regex.search(message_text):
            message_offset = int(time_offset.group(1))
            message_text = "".join(message_text.rsplit(time_offset.group(), 1))
        for prefix in VALID_PREFIXES:
            if message_text.startswith(prefix):
                tags.append(
                    {
                        "id": message_id,
                        "timestamp": datetime.fromtimestamp(
                            int(message_time), tz=timezone.utc
                        ),
                        "text": message_text.replace(prefix, "", 1).strip(),
                        "quoting": message_quoting,
                        "offset": message_offset,
                    }
                )
                break
    return tags


def sort_and_edit(tags):
    tags.sort(key=lambda i: i["timestamp"])
    remove_tags = []
    tag_ids = {}
    shifted = False
    for tag in tags:
        tag_ids[tag["id"]] = tag
        if tag["text"].startswith("edit "):
            if tag["quoting"] and tag["quoting"] in tag_ids:
                tag_ids[tag["quoting"]]["text"] = (
                    tag["text"].replace("edit ", "", 1).lstrip()
                )
            remove_tags.append(tag)
        elif tag["text"] == "ignore":
            if tag["quoting"] and tag["quoting"] in tag_ids:
                remove_tags.append(tag_ids[tag["quoting"]])
            remove_tags.append(tag)
        elif tag["text"] == "shift":
            if tag["quoting"] and tag["quoting"] in tag_ids:
                tag_ids[tag["quoting"]]["timestamp"] = tag["timestamp"]
                shifted = True
            remove_tags.append(tag)
    for tag in remove_tags:
        if tag in tags:
            tags.remove(tag)
    if shifted:
        tags.sort(key=lambda i: i["timestamp"])


def get_4chan_tags(video_attributes, tags, default_offset):
    output = f"Tags for stream {video_attributes['title']}\n\n"
    total_lines = 2
    additional_delay = 0
    for tag in tags:
        time_diff = (tag["timestamp"] - video_attributes["timestamp"]).total_seconds()
        if time_diff >= 0:
            if new_delay := add_delay_regex.fullmatch(tag["text"]):
                additional_delay += int(new_delay.group(1))
                continue
            if total_lines >= MAX_4CHAN_POST_LINES - 1:
                output += "=======================================\n"
                total_lines = 1
            tag_offset = max(
                0, int(time_diff - default_offset - tag["offset"] - additional_delay)
            )
            tag_url = f"https://youtu.be/{video_attributes['id']}?t={(tag_offset)}"
            output += f">{tag['text']}\n{tag_url}\n"
            total_lines += 2
    return output


def get_youtube_tags(video_attributes, tags, default_offset):
    output = "Timestamps for stream:\n\n"
    additional_delay = 0
    for tag in tags:
        time_diff = (tag["timestamp"] - video_attributes["timestamp"]).total_seconds()
        if time_diff >= 0:
            if new_delay := add_delay_regex.fullmatch(tag["text"]):
                additional_delay += int(new_delay.group(1))
                continue
            time_offset = max(
                0, time_diff - default_offset - tag["offset"] - additional_delay
            )
            hours = int(time_offset / 3600)
            minutes = int((time_offset % 3600) / 60)
            seconds = int(time_offset % 60)
            hour_string = f"{hours:02d}:" if hours else ""
            minute_string = f"{minutes:02d}:" if hours or minutes else ""
            second_string = f"{seconds:02d}"
            escaped_text = (
                tag["text"]
                .replace("*", "\u200B*")
                .replace("-", "\u200B-")
                .replace("_", "\u200B_")
            )
            output += f"{hour_string}{minute_string}{second_string} {escaped_text}\n"
    return output


@app.route("/", methods=["GET"])
def index():
    request_form = RequestForm()
    return render_template(
        "index.html", request_form=request_form, URL_PREFIX=URL_PREFIX
    )


@app.route("/request", methods=["POST"])
def perform_request():
    try:
        data = request.get_json()
        video_url = data.get("video_entry")
        thread_urls = data.get("threads_entry")
        default_offset = int(data.get("delay_entry"))
        logger.info(f"Performing request with parameters {repr(data)}")
        request_start = datetime.now()
        video_id = video_url.rpartition("/")[2].rpartition("?v=")[2][:11]
        video_attributes = get_video_attributes(video_id)
        tags = []
        for url in thread_urls.split("\n"):
            if id_match := re.match(r"\d+", url.rpartition("/")[2]):
                thread_id = id_match.group()
                tags.extend(parse_thread(thread_id))
        sort_and_edit(tags)
        logger.info(f"Found {len(tags)} timestamps in {datetime.now() - request_start}")
        return {
            "Success": True,
            "BreadTags": get_4chan_tags(video_attributes, tags, default_offset),
            "YoutubeTags": get_youtube_tags(video_attributes, tags, default_offset),
        }
    except Exception as e:
        logger.error("Error performing request.")
        logger.exception(e)
        return {"Success": False}
