import bleach
import logging
import re
import sqlite3
from datetime import datetime
from dateutil.parser import parse as parsedate
from flask import Flask, redirect, render_template, request, g, url_for
from flask_wtf import FlaskForm
from itertools import chain
from shlex import split
from urllib.parse import quote
from wtforms import IntegerField, RadioField, StringField, SubmitField
from wtforms.validators import Length, InputRequired, NumberRange

DATABASE = "/disk1/pyondex/database.db"
LOG_FILE = "/disk1/pyondex/server.log"
URL_PREFIX = "/pyondex"
ARCHIVE_BASE_URL = "https://warosu.org"
SELECTED_BOARD = "/vt/"
MAX_RESULTS = 100

INCLUDE_OP = 0
EXCLUDE_OP = 1
OPS_ONLY = 2


class SourceFilter(logging.Filter):
    def filter(self, record):
        record.request_source = (
            request.headers["Cf-Connecting-Ip"]
            if "Cf-Connecting-Ip" in request.headers
            else "Local"
        )
        return True


class SearchForm(FlaskForm):
    text_search = StringField("Text to find", validators=[Length(max=200)])
    subject_search = StringField("Subject", validators=[Length(max=50)])
    username_search = StringField("Username", validators=[Length(max=50)])
    tripcode_search = StringField("Tripcode", validators=[Length(max=50)])
    filename_search = StringField("Filename", validators=[Length(max=200)])
    from_date_search = StringField("From Date", validators=[Length(max=20)])
    to_date_search = StringField("To Date", validators=[Length(max=20)])
    include_ops_search = RadioField(
        "Include OPs",
        choices=[
            (INCLUDE_OP, "Show OPs"),
            (EXCLUDE_OP, "Hide OPs"),
            (OPS_ONLY, "Only Show OPs"),
        ],
        default=INCLUDE_OP,
    )
    oldest_first_search = RadioField(
        "Post Order",
        choices=[(0, "New posts first"), (1, "Old posts first")],
        default=0,
    )
    limit_search = IntegerField(
        "Limit",
        validators=[InputRequired(), NumberRange(min=1, max=MAX_RESULTS)],
        default=20,
    )
    offset_search = IntegerField(
        "Offset", validators=[InputRequired(), NumberRange(min=0)], default=0
    )
    submit = SubmitField("Search")


logger = logging.getLogger(__name__)
log_handler = logging.FileHandler(LOG_FILE, encoding="utf-8")
log_handler.setFormatter(
    logging.Formatter(
        fmt="%(asctime)s [%(levelname)s] (%(request_source)s) %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
)
log_handler.addFilter(SourceFilter())
logging.basicConfig(handlers=[log_handler], level=logging.INFO)

app = Flask(__name__)
app.config["SECRET_KEY"] = ""
app.jinja_env.add_extension("jinja2.ext.do")

or_token_regex = re.compile(r"\s*\|\s*")


def get_db():
    db = getattr(g, "_database", None)
    if db is None:
        db = g._database = sqlite3.connect(DATABASE)
        db.row_factory = sqlite3.Row
    return db


def tokenize_search(search_string, column_name):
    search_string = or_token_regex.sub("|", search_string)
    if '"' not in search_string:
        search_string = search_string.replace("'", "\\'")
    try:
        and_terms = split(search_string)
    except ValueError:
        and_terms = search_string.split(" ")
    or_terms = [term for term in and_terms if "|" in term]
    for index, group in enumerate(or_terms):
        and_terms.remove(group)
        or_terms[index] = tuple(term for term in [*set(group.split("|"))] if term)
    if not and_terms and not or_terms:
        return None, None
    and_terms = [*set(and_terms)]
    or_terms = [*set(or_terms)]
    query_text = " AND ".join([f"{column_name} LIKE ?" for term in and_terms])
    query_text += " AND " if and_terms and or_terms else ""
    or_groups = [
        f'({" OR ".join([f"{column_name} LIKE ?" for term in group])})'
        for group in or_terms
    ]
    query_text += " AND ".join(or_groups)
    query_terms = tuple(
        f'%{term.replace("^", "^^").replace("%", "^%")}%'
        for term in chain(and_terms, *or_terms)
    )
    return query_text, query_terms


def perform_search(search_params):
    db = get_db()
    db_cur = db.cursor()
    query_text = "SELECT * FROM posts WHERE "
    query_terms = []
    tokenized_searches = []
    if "text_search" in search_params and search_params["text_search"]:
        tokenized_searches.append((search_params["text_search"], "plain_text"))
    if "subject_search" in search_params and search_params["subject_search"]:
        tokenized_searches.append((search_params["subject_search"], "subject"))
    if "username_search" in search_params and search_params["username_search"]:
        tokenized_searches.append((search_params["username_search"], "name"))
    if "tripcode_search" in search_params and search_params["tripcode_search"]:
        tokenized_searches.append((search_params["tripcode_search"], "tripcode"))
    if "filename_search" in search_params and search_params["filename_search"]:
        tokenized_searches.append((search_params["filename_search"], "image_filename"))
    tokenized_search_text = []
    for search in tokenized_searches:
        new_text, new_terms = tokenize_search(*search)
        if new_text:
            tokenized_search_text.append(new_text)
            query_terms.extend(new_terms)
    if not tokenized_search_text:
        tokenized_search_text.append("plain_text LIKE '%'")
    query_text += " AND ".join(tokenized_search_text)
    after_timestamp = None
    before_timestamp = None
    if "after_date" in search_params and search_params["after_date"]:
        try:
            after_timestamp = int(
                parsedate(search_params["after_date"]).timestamp() * 1000
            )
        except:
            logger.error(
                f'Could not parse after datestamp {search_params["after_date"]}'
            )
    if "before_date" in search_params and search_params["before_date"]:
        try:
            before_timestamp = int(
                parsedate(search_params["before_date"]).timestamp() * 1000
            )
        except:
            logger.error(
                f'Could not parse before datestamp {search_params["before_date"]}'
            )
    if before_timestamp:
        if after_timestamp:
            query_text += (
                f" AND timestamp BETWEEN {after_timestamp} AND {before_timestamp}"
            )
        else:
            query_text += f" AND timestamp <= {before_timestamp}"
    elif after_timestamp:
        query_text += f" AND timestamp >= {after_timestamp}"
    if int(search_params["include_op"]) == EXCLUDE_OP:
        query_text += " AND id != thread"
    elif int(search_params["include_op"]) == OPS_ONLY:
        query_text += " AND id = thread"
    query_text += (
        f' ORDER BY id {"ASC" if int(search_params["oldest_first"]) else "DESC"}'
    )
    query_text += " LIMIT ? OFFSET ?;"
    query_terms.extend(
        (min(int(search_params["limit"]), MAX_RESULTS), search_params["offset"])
    )
    db_cur.execute(query_text, query_terms)
    results = db_cur.fetchall()
    for index, result in enumerate(results):
        result = dict(result)
        result["html_text"] = result["html_text"].replace(
            f'<a href="{SELECTED_BOARD}', f'<a href="{URL_PREFIX}{SELECTED_BOARD}'
        )
        results[index] = result
    return results


def get_form_defaults_from_url():
    form_defaults = {}
    if text_search := request.args.get("text_search"):
        form_defaults["text_search"] = text_search
    if subject_search := request.args.get("subject_search"):
        form_defaults["subject_search"] = subject_search
    if username_search := request.args.get("username_search"):
        form_defaults["username_search"] = username_search
    if tripcode_search := request.args.get("tripcode_search"):
        form_defaults["tripcode_search"] = tripcode_search
    if filename_search := request.args.get("filename_search"):
        form_defaults["filename_search"] = filename_search
    if from_date_search := request.args.get("after_date"):
        form_defaults["from_date_search"] = from_date_search
    if to_date_search := request.args.get("before_date"):
        form_defaults["to_date_search"] = to_date_search
    if include_ops_search := request.args.get("include_op"):
        form_defaults["include_ops_search"] = include_ops_search
    if oldest_first_search := request.args.get("oldest_first"):
        form_defaults["oldest_first_search"] = oldest_first_search
    if limit_search := request.args.get("limit"):
        form_defaults["limit_search"] = limit_search
    return form_defaults


def get_url_args_from_form(search_form):
    ts_hour = int(datetime.utcnow().timestamp() / 3600)
    url_arguments = {}
    if text_search := search_form.text_search.data.strip():
        url_arguments["text_search"] = text_search
    if subject_search := search_form.subject_search.data.strip():
        url_arguments["subject_search"] = subject_search
    if username_search := search_form.username_search.data.strip():
        url_arguments["username_search"] = username_search
    if tripcode_search := search_form.tripcode_search.data.strip():
        url_arguments["tripcode_search"] = tripcode_search
    if filename_search := search_form.filename_search.data.strip():
        url_arguments["filename_search"] = filename_search
    if after_date := search_form.from_date_search.data.strip():
        url_arguments["after_date"] = after_date
    if before_date := search_form.to_date_search.data.strip():
        url_arguments["before_date"] = before_date
    url_arguments["include_op"] = search_form.include_ops_search.data
    url_arguments["oldest_first"] = search_form.oldest_first_search.data
    url_arguments["limit"] = search_form.limit_search.data
    url_arguments["offset"] = search_form.offset_search.data
    url_arguments["t"] = ts_hour
    return url_arguments


def render_page(template_file, display_data=None):
    search_form = SearchForm(**get_form_defaults_from_url())
    ts_hour = int(datetime.utcnow().timestamp() / 3600)
    if request.method == "POST":
        if search_form.validate_on_submit():
            url_arguments = get_url_args_from_form(search_form)
            return redirect(URL_PREFIX + url_for("search_results", **url_arguments))
    extra_offsets = []
    if request.args.get("offset") is not None and request.args.get("limit") is not None:
        offset = int(request.args["offset"])
        limit = int(request.args["limit"])
        extra_offsets = []
        last_offset = None
        for x in range(1, 4):
            if (new_offset := offset - limit * x) >= 0:
                extra_offsets.insert(0, (f"-{limit * x}", new_offset))
                last_offset = new_offset
            else:
                if offset != 0 and last_offset != 0:
                    extra_offsets.insert(0, (f"-{offset}", 0))
                break
        for x in range(1, 4):
            extra_offsets.append((f"+{limit * x}", offset + limit * x))
    return render_template(
        template_file,
        display_data=display_data,
        search_form=search_form,
        ts_hour=ts_hour,
        bleach=bleach,
        bleach_tags=("a", "span", "br", "div"),
        bleach_attrs={"a": ("href", "title"), "span": ("class",), "div": ("class",)},
        datetime=datetime,
        extra_offsets=extra_offsets,
        URL_PREFIX=URL_PREFIX,
    )


@app.teardown_appcontext
def close_connection(exception):
    db = getattr(g, "_database", None)
    if db is not None:
        db.close()


@app.route("/", methods=["GET"])
def index():
    return render_page("index.html")


@app.route("/search", methods=["GET", "POST"])
def search_results():
    if request.method == "GET":
        try:
            logger.info(f"Running search with parameters: {request.args}")
            request.start = datetime.now()
            display_data = perform_search(request.args)
            logger.info(f"Request completed in {datetime.now() - request.start}")
        except sqlite3.OperationalError as e:
            if e.args[0] == "database is locked":
                logger.error("Database is locked.")
                display_data = {"error": "DATABASE_LOCKED"}
            else:
                raise e
    else:
        display_data = None
    return render_page("search.html", display_data)


@app.route("/vt/<link_type>/<location>", methods=["GET"])
def redirect_links(link_type="", location=""):
    return redirect(f"{ARCHIVE_BASE_URL}/vt/{link_type}/{location}")
